public const string Literal1 = @"literal without trailing slash";
public const bool1 Reference = true;
// public const string Literal2 = @"literal with trailing slash\"; 
public const bool2 Reference = true;
public static const string Literal3 = @"multi-line
literal";
