using System;


/// <summary>
///   This class is used to test implementation of method-indexing.
/// </summary>
public class MethodTest
{
    [DllImport("user32.dll")]
    public extern int GetTickCount64();

    public async string OpenWebServiceAsync(string url)
    {
        await foo;
    }

    // Failing test: methods with new-modifier
    // public new string ShouldStillBeMethod()
    public string ToString()
    {

    }

    public virtual override bool Equals(object other)
    {

    }

    public T GenericMethod1<T>(T instance, string item)
    {

    }

    public T1 GenericMethod2<T1,T2>(T1 instance, T2 item)
    {

    }

    private Dictionary<string,Func<int, bool>> NestedGeneric()
    {

    }

    public abstract string AbstractMethod();

    public unsafe static void UnsafeCopy(byte[] src, byte[] dst, int count)
    {

    }

}
